#if !defined TREEBUFFER_H
#define TREEBUFFER_H

typedef struct Node Node;
typedef struct Tree Tree;

void delete(Tree * tree);
void get(Tree * tree, Node * ancestors[]);
/* Adds a node to a old tree, results in a new tree */
Tree * add(int data, Tree * tree);
/* Creates a new tree */
Tree * empty(int history);
/* Get the data in a node */
int get_data(Node * node);


#endif
